'Extentions of Docker SDK for Python'

import io
import shlex
import tarfile


def tar_bin(filename, data):
    'Write the binary file `filename` into a tarfile'
    bytes_io = io.BytesIO()
    tar = tarfile.open(mode='w', fileobj=bytes_io)
    tarinfo = tarfile.TarInfo(filename)
    tarinfo.size = len(data)
    tar.addfile(tarinfo, io.BytesIO(data))
    tar.close()
    bytes_io.seek(0)
    return bytes_io.read()


def put_bin(container, path, data):
    'Extends docker.models.containers.Container.put_archive'
    return container.put_archive(path.parent, tar_bin(path.name, data))


def get_bin(container, path):
    'Extends docker.models.containers.Container.get_archive'
    bytes_io = io.BytesIO()
    for chunk in container.get_archive(path)[0]:
        bytes_io.write(chunk)
    bytes_io.seek(0)
    tar = tarfile.open(mode='r', fileobj=bytes_io)
    return tar.extractfile(path.name).read()


def exec_run(container, command, workdir):
    'Extends docker.models.containers.Container.exec_run'
    if command:
        return container.exec_run(f'sh -c {shlex.quote(command)}'
                                  if isinstance(command, str) else command,
                                  workdir=workdir)
    return None



'Processors'
# pylint: disable = missing-function-docstring, too-few-public-methods

from enum import Enum
from pathlib import PurePosixPath


class Processor():
    'Defines the operations of a multi-version programming language processor'

    @staticmethod
    def _get_image_with_tag(image, tag):
        return image + (f':{tag}' if tag else '')

    image = None
    workdir = PurePosixPath('/dockerjudge')
    source = None
    before_compile = None
    compile = None
    after_compile = None
    before_judge = None
    judge = None
    after_judge = None


class GCC(Processor):
    'GNU project C and C++ compiler'

    class Language(Enum):
        'C or C++'
        c = 'C'
        cpp = 'C++'

    def _get_language(self, language):
        if isinstance(language, self.Language):
            return language
        return (self.Language[language] if language in ['c', 'cpp']
                else self.Language(language) if language in ['C', 'C++']
                else self.Language.cpp)

    def __init__(self, language=None, version=None,
                 filenames=None, options=None):
        lang = self._get_language(language)
        fns = filenames or {}
        args = options or []

        self.image = self._get_image_with_tag('gcc', version)
        self.source = fns.get('src', f'a.{lang.name}')
        self.compile = ([{self.Language.c: 'gcc',
                          self.Language.cpp: 'g++'}[lang],
                         self.source]
                        + (['-o', fns['bin']]
                           if fns.get('bin') else []) + args)
        self.after_compile = ['rm', self.source]
        self.judge = f"./{fns.get('bin', 'a.out')}"


class Python(Processor):
    'CPython'

    def __init__(self, version=None):
        self.image = self._get_image_with_tag('python', version)
        self.source = '__init__.py'
        self.compile = ['python', '-m', 'compileall', '.']
        self.judge = f'python {self.source}'


class Go(Processor):
    'The Go Programming Language'

    def __init__(self, version=None, filenames=None, options=None):
        fns = filenames or {}
        args = options or []

        self.image = self._get_image_with_tag('golang', version)
        self.source = fns.get('src', 'main.go')
        self.compile = (['go', 'build']
                        + (['-o', fns['bin']]
                           if fns.get('bin') else []) + args
                        + [self.source])
        self.after_compile = ['rm', self.source]
        self.judge = f"./{fns.get('bin', 'main')}"


class OpenJDK(Processor):
    'Open Java Development Kit'

    def __init__(self, version=None):
        self.image = self._get_image_with_tag('openjdk', version)
        self.source = 'Main.java'
        self.compile = ['javac', self.source]
        self.after_compile = ['rm', self.source]
        self.judge = 'java Main'


class Node(Processor):
    'Node.js®'

    def __init__(self, version=None):
        self.image = self._get_image_with_tag('node', version)
        self.source = 'index.js'
        self.compile = ['node', '-c', self.source]
        self.judge = f'node {self.source}'



import codecs
import os.path
import re

import setuptools


def read(path):
    with codecs.open(os.path.join(os.path.abspath(os.path.dirname(__file__)),
                                  path)) as fp:
        return fp.read()


def get_version(file_path):
    versions = re.findall(r"^__version__ = ['\"]([^'\"]*)['\"]",
                          read(file_path), re.M)
    if versions:
        return versions[-1]
    raise RuntimeError('Unable to find version string.')


setuptools.setup(
    name='dockerjudge',
    version=get_version('dockerjudge/__init__.py'),
    author='汪心禾',
    author_email='wangxinhe06@gmail.com',
    description='A Docker Based Online Judge Engine',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    url='https://github.com/piterator-org/dockerjudge',
    project_urls={
        "Bug Tracker": "https://github.com/piterator-org/dockerjudge/issues",
        "Documentation": "https://github.com/piterator-org/dockerjudge#readme",
        "Source Code":
            "https://github.com/piterator-org/dockerjudge/tree/master",
    },
    packages=['dockerjudge'],
    install_requires=[
        'docker[tls]>=3.7'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
)



'dockerjudge - A Docker Based Online Judge Engine'

from math import ceil
from pathlib import PurePosixPath

import docker

from .dockerpy import exec_run, put_bin
from .status import Status
from . import test_case
from .thread import Thread

__version__ = '1.1.0'


def judge(processor, source, tests, config=None,
          client=docker.from_env()):
    'Main function'
    container = client.containers.run(processor.image, detach=True,
                                      network_disabled=True, tty=True)
    try:
        return run(container, processor, source, tests, config)
    finally:
        container.remove(force=True)


def compile_source_code(container, processor, source, config):
    'Compile the source file'
    container.exec_run(f"mkdir -p {processor.workdir}/0")
    put_bin(
        container,
        PurePosixPath(f'{processor.workdir}/0/{processor.source}'),
        source
    )

    exec_run(container, processor.before_compile, f'{processor.workdir}/0')
    exec_result = container.exec_run(processor.compile,
                                     workdir=f'{processor.workdir}/0')
    if 'compile' in config.get('callback', {}):
        config['callback']['compile'](exec_result.exit_code,
                                      exec_result.output)
    exec_run(container, processor.after_compile, f'{processor.workdir}/0')
    return exec_result


def judge_test_cases(container, processor, tests, config):
    'Judge test cases'
    threads_limit = config.get('threads') or len(tests)
    res = []
    for i in range(ceil(len(tests) / threads_limit)):
        threads = []
        for j in range(i * threads_limit,
                       min((i + 1) * threads_limit, len(tests))):
            threads.append(
                Thread(
                    target=test_case.__init__,
                    args=(container, processor, j + 1, tests[j], config),
                    callback=config.get('callback', {}).get('judge')
                )
            )
            threads[-1].start()
        for thread in threads:
            thread.join()
            res.append(thread.return_value)
    return res


def run(container, processor, source, tests, config=None):
    'Compile and judge'
    config = config or {}
    exec_result = compile_source_code(container, processor, source, config)
    if exec_result.exit_code:
        return [[[Status.CE, (None, None), .0]] * len(tests),
                exec_result.output]

    res = judge_test_cases(container, processor, tests, config)
    return [res, exec_result.output]



'Extentions for Lib/threading.py'

import threading

from .status import Status


class Thread(threading.Thread):
    'Subclass of threading.Thread with return value and callback'

    def __init__(self, callback, *args, **kwargs):
        self.return_value = (Status.UE, .0)
        self._callback = callback
        super().__init__(*args, **kwargs)

    def run(self):
        try:
            self.return_value = self._target(*self._args, **self._kwargs)
        finally:
            try:
                self._callback(self._args[2] - 1, *self.return_value)
            except TypeError:
                pass
            finally:
                del self._target, self._args, self._kwargs, self._callback
